/*		Character grid hypertext object
**		===============================
**              Converted for use with tkWWW
*/

#include <assert.h>
#include <ctype.h>
#include <tcp.h>
#include "HTUtils.h"
#include "HTString.h"
#include "GridText.h"
#include "HTFont.h"


struct _HTStream {			/* only know it as object */
    CONST HTStreamClass *	isa;
    /* ... */
};

PRIVATE int HTScreenWidth = 80;

#define MAX_LINE	HTScreenWidth+24 /* No point in accumulating more */
#ifdef THINK_C
#define LOADED_LIMIT 3			/* For now, save last two texts */
#else
#define LOADED_LIMIT 6			/* For now, save last five texts */
#endif


/*	From default style sheet:
*/
extern HTStyleSheet * styleSheet;	/* Default or overridden */

/*	Exports
*/ 
PUBLIC HText * HTMainText = 0;		/* Equivalent of main window */
PUBLIC HTParentAnchor * HTMainAnchor = 0;	/* Anchor for HTMainText */

#define LINE_SIZE(l) (sizeof(HTLine)+(l))	/* allow for terminator */

PRIVATE HTStyle default_style =
	{ 0,  "(Unstyled)", "",
	HT_FONT, 1.0, HT_BLACK,		0, 0,
	0, 0, 0, HT_LEFT,		1, 0,	0, 
	NO, NO, 0, 0,			0 };	

PRIVATE void blank_lines PARAMS ((HText *, int));

/*			Creation Method
**			---------------
*/
PUBLIC HText *	HText_new ARGS1(HTParentAnchor *,anchor)
{
    HText * self = (HText *) malloc(sizeof(*self));
    if (!self) return self;

    self->line_position = 0;
    self->offset = 0;
    self->lines = self->chars = 0;
    self->title = 0;
    self->first_anchor = self->last_anchor = 0;
    self->style = &default_style;
    self->line_begin_style = &default_style;

    self->node_anchor = anchor;
    self->last_anchor_number = 0;	/* Numbering of them for references */
    self->source = NULL;
    self->output = HTChunkCreate(1024);
    self->current_line = HTChunkCreate(1024);
    self->blank_space = 0;
    self->in_command_mode = YES;

    HTAnchor_setDocument(anchor, (HyperDoc *)self);

    HTMainText = self;
    HTMainAnchor = anchor;
    return self;
}

/*			Creation Method 2
**			---------------
**
**	Stream is assumed open and left open.
*/
PUBLIC HText *	HText_new2 ARGS2(
		HTParentAnchor *,	anchor,
		HTStream*,		stream)

{
    HText * this = HText_new(anchor);
        
    if (stream) {
        this->target = stream;
	this->targetClass = *stream->isa;	/* copy action procedures */
    }    
    return this;
}


/*	Free Entire Text
**	----------------
*/
PUBLIC void 	HText_free ARGS1(HText *,self)
{
    HTAnchor_setDocument(self->node_anchor, (HyperDoc *)0);
    while(self->first_anchor) {		/* Free off anchor array */
        TextAnchor * l = self->first_anchor;
	self->first_anchor = l->next;
	HTChunkFree(l->href);
	free(l);
    }
    HTChunkClear(self->current_line);
    HTChunkClear(self->output);
    if (self->source)
      free(self->source);
    free(self);
}

/*			Object Building methods
**			-----------------------
**
**	These are used by a parser to build the text in an object
*/

PRIVATE void add_text ARGS3(HTChunk *, chunk, int, anchor_num, 
			    HTStyle *, style) 
{
  if (anchor_num == 0) {
    HTChunkPuts(chunk, "tkW3HtBuf ");
    HTChunkPuts(chunk, style->name);
    HTChunkPuts(chunk, " \"");
  } else {
    char marker[100];
    sprintf(marker, "tkW3HtAnc %d \"%s\" \"",
	    anchor_num, style->name);
    HTChunkPuts(chunk, marker);
  }
}

PUBLIC void HText_beginAppend ARGS1(HText *,text)
{
  text->output->size = 0;
  text->permissible_split = 0;
  text->split_line_position = 0;
  text->in_line_1 = YES;
  text->current_anchor = 0;
  text->line_begin_anchor = 0;

  HTChunkPuts(text->output, "tkW3HtBeginDoc\n");
  text->in_command_mode = YES;
}


/*	Add a new line of text
**	----------------------
**
** On entry,
**
**	split	is zero for newline function, else number of characters
**		before split.
** On exit,
**		A new line has been made, justified according to the
**		current style. Text after the split (if split nonzero)
**		is taken over onto the next line.
*/

#define new_line(text) split_line(text, 0)

PRIVATE void split_line ARGS2(HText *,text, int,split)
{
  HTStyle *style = text->style;
  char *data;
  int spare, i;
  int indent = text->in_line_1 ? text->style->indent1st
                             : text->style->leftIndent;
/*	Align left, right or center
*/

  spare =  HTScreenWidth -
    style->rightIndent + style->leftIndent -
    text->line_position;	/* @@ first line indent */
		
  switch (style->alignment) {
  case HT_CENTER :
    text->offset += (indent + spare/2);
    break;
  case HT_RIGHT :
    text->offset += (indent + spare);
    break;
  case HT_LEFT :
  case HT_JUSTIFY :		/* Not implemented */
  default:
    text->offset += indent;
    break;
  } /* switch */

  if (text->line_position > 0) {
    if (text->offset > 0) {
      HTChunkPuts(text->output, "tkW3HtBuf offset \"");
      for (i=0; i < text->offset; i++)
	HTChunkPutc(text->output, ' ');
      HTChunkPuts(text->output, "\"\n");
    }
    text->blank_space = 0;
  } else {
    if (text->blank_space == 0) {
      HTChunkPuts(text->output, "tkW3HtBuf blankline \"\n\"\n");
      text->blank_space++;
    }
  }
  HTChunkTerminate(text->current_line);

  data = text->current_line->data;
  /* Split at required point */
  if (split > 0) {
    char * p;
    char * copy_string;

    for (i=0; i< split; i++) 
      HTChunkPutc(text->output, data[i]);


    for (p = &data[split]; *p; p++) {
      if (*p != ' ') break;
      text->split_line_position++;
    }
    
/*    if (*p) { */
      copy_string = malloc(strlen(p) + 1);
      strcpy(copy_string, p); 

    text->current_line->size = 0;
    HTChunkPuts(text->output, "\n\"\n");
    add_text(text->current_line, text->current_anchor, text->style);    
    HTChunkPuts(text->current_line, copy_string);
      free(copy_string);
      text->line_position -= text->split_line_position;
      text->in_command_mode = NO;
 /*   } else {
      if (!text->in_command_mode)
	HTChunkPuts(text->output, "\n\"\n");
      text->in_command_mode = YES;
      text->current_line->size = 0;
      text->line_position = 0;
    } */
  } else {
    HTChunkPuts(text->output, data);
    text->current_line->size = 0;

    if (text->line_position > 0) {
      if (text->in_command_mode)
	add_text(text->output, text->current_anchor, text->style); 
      HTChunkPuts(text->output, "\n\"\n");
    } else {
      if (!text->in_command_mode) 
	HTChunkPuts(text->output, "\n\"\n");
    }

    text->line_position = 0;
    text->in_command_mode = YES;
  }

  text->line_begin_style = text->style;
  text->line_begin_anchor = text->current_anchor;
  text->offset = 0;
  text->permissible_split = 0;
  text->split_line_position = 0;
  text->in_line_1 = NO;
}

PRIVATE void blank_lines ARGS2(HText *, text, int, newlines)
{
  int i;
  new_line(text);
  if (newlines > text->blank_space) {
    HTChunkPuts(text->output, "tkW3HtBuf blankline \"");
    for (i=0; i < newlines - text->blank_space; i++) {
      HTChunkPutc(text->output, '\n');
    }
    HTChunkPuts(text->output, "\"\n");
    text->blank_space = newlines;
  }
  text->in_line_1 = YES;
}
  
/*	New paragraph in current style
**	------------------------------
** See also: setStyle.
*/

PUBLIC void HText_appendParagraph ARGS1(HText *,text)
{
  int after = text->style->spaceAfter;
  int before = text->style->spaceBefore;
  blank_lines(text, after>before ? after : before);
}

PUBLIC void HText_appendBullet ARGS3(HText *, text, CONST char *, type,
				     BOOL, in_line_1) {
  new_line(text);
  text->in_line_1 = in_line_1;
  if (!text->in_command_mode) {
    HTChunkPuts(text->current_line, "\"\n");
    text->in_command_mode = YES;
  } 

  HTChunkPuts(text->current_line, "tkW3HtBul ");
  HTChunkPuts(text->current_line, text->style->name);
  HTChunkPuts(text->current_line, " ");
  HTChunkPuts(text->current_line, type);
  HTChunkPuts(text->current_line, "\n");
}
  
PUBLIC void HText_setSource ARGS2(HText *, text, CONST char *, string) {
  if (text->source) 
    free(text->source);
  text->source = (char *) malloc(strlen(string) +1);
  strcpy (text->source, string);
}

/*	Set Style
**	---------
**
**	Does not filter unnecessary style changes.
*/
PUBLIC void HText_setStyle ARGS2(HText *,text, HTStyle *,style)
{
  int after, before;

  /* Just in case some goes wrong, return if the style is NULL */
  if (!style) return;	

  if (text->style != &default_style) {
  /* Add the correct amount of space between styles */
    after = text->style->spaceAfter;
    before = style->spaceBefore;
    blank_lines (text, after>before ? after : before);
  }

  /* Do the change */
  text->style = style;
}

/*	Append a character to the text object
**	-------------------------------------
*/
PUBLIC void HText_appendCharacter ARGS2(HText *,text, char,ch)
{
  HTTabStop * tab;
  HTStyle *style = text->style;
  int indent = text->in_line_1 ? style->indent1st : style->leftIndent;

  int target, here;	/* Where to tab to */
  if (text->in_command_mode) {
    add_text(text->current_line, text->current_anchor, text->style);
    text->in_command_mode = NO;
  }
  switch (ch) {
  case '{': /* add an escape before braces */
  case '}':
  case '[':
  case ']':
  case '\"':
  case '\\':
  case '$':
    HTChunkPutc(text->current_line, '\\');
    break;
  case '\n': /* explicit newline */
    new_line(text);
    text->in_line_1 = YES;
    return;

  case '\t': /* Handle tab */
    here = text->line_position + text->offset + indent;
    if (style->tabs) {	/* Use tab table */
      for (tab = style->tabs; tab->position <= here; tab++)
	if (!tab->position) {
	  new_line(text);
	  return;
	}
      target = tab->position;
    } else if (text->in_line_1) {	/* Use 2nd indent */
      if (here >= style->leftIndent) {
	new_line(text); /* wrap */
	return;
      } else {
	target = style->leftIndent;
      }
    } else {		/* Default tabs align with left indent mod 8 */
#ifdef DEFAULT_TABS_8
      target = ((text->line_position + 8) & (-8)) + style->leftIndent;
#else
      new_line(text);
      return;
#endif
    }
    if (target > HTScreenWidth - style->rightIndent) {
      new_line(text);
      return;
    } else {
      /* Actually add the tab */
      text->permissible_split = text->current_line->size; /* Can split here */
      text->split_line_position = text->line_position;

      /* If there are no characters in the line, add shift the offset of the
       * whole line.
       */
      if (text->line_position == 0) 
	text->offset += (target - here);

      /* Add spaces so that the tab reaches the target */
      for(; here<target; here++) {
	HTChunkPutc(text->current_line, ' ');
	text->line_position++;
      }
      return;
    }
    break;

  case ' ':
    text->permissible_split = text->current_line->size;
    text->split_line_position = text->line_position;
    break;
  }

  if (indent + text->offset + text->line_position + style->rightIndent
      >= HTScreenWidth && style->wordWrap) {
      split_line(text, text->permissible_split);
      if (ch==' ') return;	/* Ignore space causing split */
    }

  if (ch == HT_NON_BREAK_SPACE)
    ch = ' ';
    
  HTChunkPutc(text->current_line, 
	      style->font & HT_CAPITALS ? TOUPPER(ch) : ch);
  text->line_position++;

  if (style->font & HT_DOUBLE)
    HText_appendCharacter(text, HT_NON_BREAK_SPACE);
}

/*		Anchor handling
**		---------------
*/
/*	Start an anchor field
*/

PUBLIC void HText_beginAnchor ARGS2(HText *,text, HTChildAnchor *,anc)
{
  HText_addAnchor(text, anc);
  if (!text->in_command_mode) {
    HTChunkPuts(text->current_line, "\"\n");
    text->in_command_mode = YES;
  }
}

PUBLIC void HText_setAnchorHref ARGS2(HText *, text, CONST char *, href)
{
  if (href) {
    HTChunkClear(text->last_anchor->href);
    HTChunkPuts(text->last_anchor->href, href);
    HTChunkTerminate(text->last_anchor->href);
  }
}

PUBLIC int HText_addAnchor ARGS2(HText *,text, HTChildAnchor *,anc)
{
  TextAnchor * a = (TextAnchor *) malloc(sizeof(*a));
  if (a == NULL) outofmem(__FILE__, "HText_beginAnchor");
  a->extent = 0;

  if (text->last_anchor) 
    text->last_anchor->next = a;
  else 
    text->first_anchor = a;

  a->next = 0;
  a->anchor = anc;
  text->last_anchor = a;
  a->href = HTChunkCreate(1024);
  HTChunkTerminate(a->href);

  text->current_anchor = a->number = ++(text->last_anchor_number);
  return a->number;
}

PUBLIC void HText_endAnchor ARGS1(HText *,text)
{
  text->current_anchor = 0;
  if (!text->in_command_mode) {
    HTChunkPuts(text->current_line, "\"\n");
    text->in_command_mode = YES;
  }
}

PUBLIC void HText_appendText ARGS2(HText *,text, CONST char *,str)
{
  CONST char * p;
  for(p=str; *p; p++)
    HText_appendCharacter(text, *p);
}


PUBLIC void HText_endAppend ARGS1(HText *,text)
{
  char marker[100];
  new_line(text);
  HTChunkPuts(text->output, "tkW3HtEndDoc\n");
  sprintf(marker, "tkW3HtBindAnc %d\n", 
	  text->last_anchor_number); 
  HTChunkPuts(text->output, marker);
  HTChunkTerminate(text->output);
}



/* 	Dump diagnostics to stderr
*/
PUBLIC void HText_dump ARGS1(HText *,text)
{
  fprintf(stderr, "HText: Dump called\n");
}
	

/*	Return the anchor associated with this node
*/
PUBLIC HTParentAnchor * HText_nodeAnchor ARGS1(HText *,text)
{
  return text->node_anchor;
}

/*				GridText specials
**				=================
*/
/*	Return the anchor with index N
**
**	The index corresponds to the number we print in the anchor.
*/
PUBLIC HTChildAnchor * HText_childNumber ARGS2(HText *,text, int,number)
{
  TextAnchor * a;
  for (a = text->first_anchor; a; a = a->next) 
    if (a->number == number) return a->anchor;
  return (HTChildAnchor *)0;	/* Fail */
}

PUBLIC CONST char * HText_getAnchorHref ARGS2(HText *,text, int,number)
{
  TextAnchor * a;
  for (a = text->first_anchor; a; a = a->next) 
    if (a->number == number) 
      return a->href->data;
  return (char *)0;	/* Fail */
}

PUBLIC int HText_sourceAnchors ARGS1(HText *,text)
{
  return text->last_anchor_number;
}

PUBLIC CONST char *HText_getSource ARGS1(HText *,text)
{
  if (text->source)
    return text->source;
  else
    return "";
}

/*		Browsing functions
**		==================
*/

PUBLIC BOOL HText_select ARGS1(HText *,text)
{
  if (text != HTMainText) {
    HTMainText = text;
    HTMainAnchor = text->node_anchor;
  }
  return YES;
}

PUBLIC BOOL HText_selectAnchor ARGS2(HText *,text, HTChildAnchor *,anchor)
{
  return YES;
}
 

/*		Editing functions		- NOT IMPLEMENTED
**		=================
**
**	These are called from the application. There are many more functions
**	not included here from the orginal text object.
*/

/*	Style handling:
*/
/*	Apply this style to the selection
*/
PUBLIC void HText_applyStyle ARGS2(HText *, me, HTStyle *,style)
{
}


/*	Update all text with changed style.
*/
PUBLIC void HText_updateStyle ARGS2(HText *, me, HTStyle *,style)
{
}


/*	Return style of  selection
*/
PUBLIC HTStyle * HText_selectionStyle ARGS2(
	HText *,me,
	HTStyleSheet *,sheet)
{
    return 0;
}


/*	Paste in styled text
*/
PUBLIC void HText_replaceSel ARGS3(
	HText *,me,
	CONST char *,aString, 
	HTStyle *,aStyle)
{
}


/*	Apply this style to the selection and all similarly formatted text
**	(style recovery only)
*/
PUBLIC void HTextApplyToSimilar ARGS2(HText *,me, HTStyle *,style)
{
    
}

 
/*	Select the first unstyled run.
**	(style recovery only)
*/
PUBLIC void HTextSelectUnstyled ARGS2(HText *,me, HTStyleSheet *,sheet)
{
}


/*	Anchor handling:
*/
PUBLIC void		HText_unlinkSelection ARGS1(HText *,me)
{    
}

PUBLIC HTAnchor *	HText_referenceSelected ARGS1(HText *,me)
{
     return 0;   
}

PUBLIC HTAnchor *	HText_linkSelTo ARGS2(HText *,me, HTAnchor *,anchor)
{
    return 0;
}


