## bookmarks.tcl tkWWW bookmarks facility
## ==============
## Copyright (C) 1992 Joseph Wang
## See the file COPYRIGHT for conditions

## This file contains code to implement the bookmarks feature of tkWWW
## Note that the file which tkWWW uses to load and save bookmarks is a
## valid HTML file.  That is to say, you can save some bookmarks into a
## file and then view that file using tkWWW.

## Some of the procedures in this file refer directly to tkWWW widgets.
## (i.e. they add or remove entries from a menu and refer directly to that
## menu).  That's probably bad abstraction.

## The bookmarks facility is quite rudimentary

## Conventions:
##   all global visible objects in this file should begin with 
##   tkWWWBookmarks

## Global Variables

# tkWWWBookmarksList contains bookmarks in 
# {{title address} {title address}} format
set tkWWWBookmarksList {}

proc tkWWWBookmarksInitialize {} {
}

# tkWWWBookmarksDefaultFile is the default file which the bookmarks are
# saved in
set tkWWWBookmarksDefaultFile "bookmarks.html" 

## tkWWWBookmarksAdd: Add a bookmark
## -----------
## Arguments
##   title - Title of the bookmark
##   address - Address bookmark refers to
##
## If called with no arguments it sets title and address to current values
## No return value
## 
## This procedure adds a bookmark to the list of bookmarks
##

proc tkWWWBookmarksAdd {{title ""} {address ""}} {
    global tkWWWBookmarksList
    if {$title == "" && $address == ""} {
	set address [tkWWWNavigateGetAddress]
	set title [tkWWWNavigateGetTitle]
    }
    lappend tkWWWBookmarksList [list $title $address]
    tkWWWMenuMakeNodeMenu .menu.bookmarks.m.bookmarks $tkWWWBookmarksList
    tkWWWConfigDisplay
}

## tkWWWBookmarksClear: clear all bookmarks
## -----------
## No arguments
## No return value
##
## This procedure clears all of the bookmarks 
##

proc tkWWWBookmarksClear {} {
    global tkWWWBookmarksList
    set tkWWWBookmarksList {}
}

proc tkWWWBookmarksDelete {} {
    tkWWWHelpNotAvailable
}

## tkWWWBookmarksLoadDialog: Brings up file dialog to load bookmarks
## -----------
## Arguments:
##   clearentries - determines if previous bookmarks should be removed
## No return value
##
## This procedure pops up a file dialog by which the user can choose a 
## file to load bookmarks from.  If clearentries is non-zero, tkWWW
## will forget about the old bookmarks, when it loads the new ones

proc tkWWWBookmarksLoadDialog {clearentries} {
# $clearentries is expanded when this procedure is called
# xfFSPath and xfFSName are not

   FSBox {Load Bookmarks} "bookmarks.html" \
	"tkWWWBookmarksLoad \$fsBox(path)/\$fsBox(name) $clearentries"
   DLG:show . .fsBox
}

proc tkWWWBookmarksSaveDialog {} {
    FSBox {Save Bookmarks} "bookmarks.html" \
         {tkWWWBookmarksSave $fsBox(path)/$fsBox(name)}
    DLG:show . .fsBox
}

# Load bookmarks from file
# Clear entries if clearentries is true

proc tkWWWBookmarksLoad {filename clearentries} {
    global tkWWWBookmarksList
    set file {}

# open the file
# trap for errors
    if {[catch {open $filename r} file]} {
	tkWWWOutputError $file
	return
    }

# clear the bookmarks
    if {$clearentries} {
	tkWWWBookmarksClear
    }

# while there is stuff in the file
    while {[gets $file string] != -1} { 
# the line contains a hypertext anchor load it
	if {[regexp {HREF="(.*)">(.*)</A>} $string {} address title]} {
	    lappend tkWWWBookmarksList [list $title $address]
	}
    }
    close $file
    .menu.bookmarks.m entryconfigure 0 -state normal
    tkWWWMenuMakeNodeMenu .menu.bookmarks.m.bookmarks $tkWWWBookmarksList
}

# save bookmarks in "filename"
proc tkWWWBookmarksSave {filename} {
    global tkWWWBookmarksList

    if {[catch {open $filename w} file]} {
	tkWWWOutputError $file
	return
    }

    puts $file "<TITLE>tkWWW Bookmarks File</TITLE>"
    puts $file "<UL>"

    set i 0
    foreach item $tkWWWBookmarksList {
	puts $file \
	    "<LI><A NAME=\"bmark$i\" HREF=\"[lindex $item 1]\">[lindex $item 0]</A>"
        incr i
    }
    puts $file "</UL>"
    close $file
}
