/**/## config.tcl: configuration file for tkWWW  user interface
/**/## ==============
/**/## Copyright (C) 1992 Joseph Wang
/**/## See the file COPYRIGHT for conditions

/**/## Conventions:
/**/##   This is the only file the user should have to edit

/**/## *************************
/**/## This file contains user definable parameters
/**/## *************************

/**/## *************************
/**/## The home page and the start page
/**/## *************************

/**/## Get it from the environment variable WWW_HOME if possible

if {[catch {set env(WWW_HOME)} home_page] != 0} {
    set tkWWWHomePage CPP_TK_WWW_HOME_PAGE
    set tkWWWStartPage CPP_TK_WWW_START_PAGE
} {
    set tkWWWHomePage $home_page
    set tkWWWStartPage $home_page
}

    set tkWWWConfigHelpRoot file:CPP_TK_WWW_SOURCE_PATH/help

/**/## The programs that tkWWW uses
set tkWWWConfigEditor emacs
set tkWWWConfigTerminal xterm
set tkWWWConfigDviViewer xdvi
set tkWWWConfigImageDirectView 0
set tkWWWConfigImageViewer xli
set tkWWWConfigVideoViewer mpeg_play
set tkWWWConfigPostscriptViewer ghostview
set tkWWWConfigAudioViewer showaudio
set tkWWWConfigPrintCommand lpr

/**/## The backend command tkWWW uses
set tkWWWServerCommand tkWWW.server

/**/## The mail command
proc tkWWWConfigSendMail {to subject body} {
#if CPP_TK_WWW_USE_MH_MAIL
    set file [open "| mhmail $to -subject \"$subject\"" w+]
#else
    set file [open "| Mail -s \"$subject\" $to" w+]
#endif
    puts $file $body
    close $file
}

/**/## Configure the menus
set tkWWWConfigMenus {
    {{file "File" 0} {
	{"New" 	{tkWWWEditClear .f.msg} 0}
	{"Open File..." tkWWWFileLoadSource 0 }
	{"Open Link..." {DLG:show . .goto_dialog}}
	{"Generate Source" tkWWWEditGenerate}
	{"Save Text..." tkWWWFileSaveText 0 }
	{"Reload page" tkWWWNavigateReload}
	{}
	{"Find Keyword..." {DLG:show . .find_dialog} 0 "<Meta-f>"}
	{"Clone Current Window" tkWWWNavigateClone 1}
	{} 
	{"Print" {DLG:show . .print_dialog} 0 }
	{"Mail" 
	    {DLG:set_entry_value .mail_dialog 2 [tkWWWNavigateGetTitle];
	     DLG:show . .mail_dialog} }
	{}
	{"Close" tkWWWFileQueryCloseWindow 0}
	{"Exit" tkWWWFileQueryExit 0}
    }}
    {{edit "Edit" 0} {
	{"Cut" {tkWWWEditCut .f.msg} 2 "<Shift-Delete>"}
        {"Copy" {tkWWWEditCopy .f.msg} 0 "<Control-Insert>"} 
	{"Paste" {tkWWWEditPaste .f.msg} 0 "<Shift-Insert>"}
	{}
	{"Delete" {tkWWWEditDelete .f.msg} 0}
	{}
	{"Select All" {tkWWWEditSelectAll .f.msg} {} }
	{"Deselect All" {tkWWWEditDeselectAll .f.msg} {} }
	{}
	{"Insert List Item" {tkWWWEditBeginListItem .f.msg} {} "<Meta-l>"}
	{"Insert Paragraph Break" {tkWWWEditInsertBullet .f.msg P} {} "<Meta-p>"}
	{"Create Link ..." {tkWWWEditBeginAnchor .f.msg} {} "<Meta-a>"}
	{"Change title ..." {tkWWWEditQueryChangeTitle .f.msg} {} "<Meta-t>"}
    }}
    {{option "Option" 0} {
	{"Change Fonts ..." {DLG:show . .font_dialog}}
    }}	
    {{style "Styles" 0} {
	{"Normal" "tkWWWEditClearTagsFromSelected .f.msg"}
	{"Headings =>" "cascade" 
	    {{"Heading1" "tkWWWEditTagSelected .f.msg Heading1" "" "<Meta-1>"}
	     {"Heading2" "tkWWWEditTagSelected .f.msg Heading2" "" "<Meta-2>"}
	     {"Heading3" "tkWWWEditTagSelected .f.msg Heading3" "" "<Meta-3>"}
	     {"Heading4" "tkWWWEditTagSelected .f.msg Heading4" "" "<Meta-4>"}
	     {"Heading5" "tkWWWEditTagSelected .f.msg Heading5" "" "<Meta-5>"}
	     {"Heading6" "tkWWWEditTagSelected .f.msg Heading6" "" "<Meta-6>"}
	     {"Heading7" "tkWWWEditTagSelected .f.msg Heading7" "" "<Meta-7>"}
	    }}
	{"List" "tkWWWEditTagSelected .f.msg List"}
	{"Address" "tkWWWEditTagSelected .f.msg Address"}
	{"Preformatted" "tkWWWEditTagSelected .f.msg Preformatted"}
    }}
    {{history "History" 0} {
	{"Backtrack" "tkWWWNavigateBacktrack" 0}
	{"Recall =>" cascade_external history}
    }}
    {{bookmarks "Bookmarks" 0} {
	{"Goto Bookmark =>" cascade_external bookmarks}
	{}
	{"Add Bookmark" tkWWWBookmarksAdd 0}
	{"Delete Bookmark" tkWWWBookmarksDelete 0}
	{"Load New Bookmarks..." {tkWWWBookmarksLoadDialog 1} 5}
	{"Append Bookmarks..." {tkWWWBookmarksLoadDialog 0} 1}
	{"Save Bookmarks..." {tkWWWBookmarksSaveDialog} 0}
    }}
    {{annotate "Annotate" 0} {
	{"Add Annotation..." tkWWWAnnotateAdd}
	{"Edit Annotation..." tkWWWAnnotateEdit}
	{}
	{"Goto Annotation =>" cascade_external annotate}
    }}
    {{documents "Documents" 0 goto} {
	{"Web Overview" 
	    "http://info.cern.ch./hypertext/WWW/LineMode/Defaults/default.html"
	     0 	}
	{"Web Project" "http://info.cern.ch/hypertext/WWW/TheProject.html" 5}
	{"Web Servers" 
	    "http://info.cern.ch/hypertext/DataSources/WWW/Servers.html" 5 }
	{"Information by subject"
	    "http://info.cern.ch/hypertext/DataSources/bySubject/Overview.html"
             0}
	{"Data Source by type of Service"
	    "http://info.cern.ch/hypertext/DataSources/ByAccess.html"
             0}
	{"InterNIC Info Source"
	    "gopher://is.internic.net:70/11/infosource" 7}
	{}
	{"WAIS Interface"
	    "http://info.cern.ch./hypertext/DataSources/WAIS/ByHost.html" 0}
	{"Gopher Interface" cascade_goto {
	    { "Gopherspace Overview" 
		"gopher://gopher.micro.umn.edu:70/11/Other%20Gopher%20and%20Information%20Servers" 0}
	    { "Veronica Interface"
		"gopher://hunter.cs.unr.edu:70/11/veronica" 0}
	    {}
	    { "NCSA Gopher" "gopher://gopher.ncsa.uiuc.edu:70/1" 0 }
	    { "CTC Gopher" "gopher://gopher.tc.cornell.edu:70/1" 0 }
	    { "PSC Gopher" "gopher://pscuxb.psc.edu:1563/1" 0 }
	    { "SDSC Gopher" "gopher://gopher.sdsc.edu:70/1" 0 }
	    {}
	    { "Original (UMN) Gopher" "gopher://gopher.micro.umn.edu:70/1" 0 }
	    { "UIUC Gopher" "gopher://gopher.uiuc.edu:70/1" 1 }
	    { "UIUC Weather Machine" "gopher://wx.atmos.uiuc.edu:70/1" 6}
	    { "USC Math Gopher" "gopher://bigcheese.math.scarolina.edu:70/1" 4 }
	    { "Wiretap Archives" "gopher://wiretap.spies.com/" 8}
	} 0}
	{"FTP Interface"
	    "http://hoohoo.ncsa.uiuc.edu:80/ftp-interface.html" 0}
	{"Usenet Interface"
	    "http://info.cern.ch/hypertext/DataSources/News/Groups/Overview.html" 0}
	{"Hyper-G Interface"
	    "http://iicm.tu-graz.ac.at:80/ROOT"}
	{"Archie Interface" 
	    "http://hoohoo.ncsa.uiuc.edu:1235/archie/index"
	    0 }
	{"HyTelnet Interface" "http://info.cern.ch.:8002/w3start.txt" 0 }
	{"Techinfo Interface" "http://info.cern.ch:8004/" 0 }
	{"Finger Interface"
	    "http://cs.indiana.edu/finger/gateway"}
	{"X.500 Interface" "gopher://umich.edu:7777/" 0}
	{"Whois Interface" "gopher://gopher.tc.umn.edu:70/11/Phone%20Books/WHOIS%20Searches" 0}
	{}
	{ "NCSA Home Page" "http://hoohoo.ncsa.uiuc.edu/welcome.html" 0}
	{ "CERN Home Page" "http://info.cern.ch/" 0}
	{"Ohio State Home Page"
	    "http://www.cis.ohio-state.edu:80/hypertext/information/information.html" 0}
	{"SLAC Home Page" "http://slacvm.slac.stanford.edu./FIND/SLAC.HTML" 1 }
	{ "HEP Overview"
	    "http://info.cern.ch/hypertext/DataSources/bySubject/Physics/HEP.html" 1}
	{}
	{ "SIGGRAPH Bibliography" "http://iicm.tu-graz.ac.at:80/CSIGGRAPHbib" 9}
	{ "Cornell Law School"
	    "http://fatty.law.cornell.edu.:80/usr2/wwwtext/lii.table.html" 3}
	{"Postmodern Culture" 
	    "gopher://una.hh.lib.umich.edu:70/11/journals/pmc" 4}
	{"Los Alamos Physics Papers"
	    "gopher://mentor.lanl.gov:70/1" }
    }}
    {{manuals "Manuals" 0 goto } {
	{ "SunOS 4.1.1 Man Pages"  
	    "http://www.cis.ohio-state.edu:80/man/sunos.top.html" 0}
	{ "HP/UX 8.0.7 Man Pages"
	    "http://www.cis.ohio-state.edu:80/man/hpux.top.html" 0 }
	{ "Ultrix 4.2a Man Pages"
	    "http://www.cis.ohio-state.edu:80/man/ultrix.top.html" 0 }
	{ "IRIX 4.0 Man Pages"
	    "http://wintermute.ncsa.uiuc.edu:1235/man/index" 0 }
	{ "AIX 3.2 Man Pages"
	    "http://hoohoo.ncsa.uiuc.edu:1235/man/index" 0 }
	{ "Pyramid OSx Man Pages"
	    "http://www.cis.ohio-state.edu:80/man/pyramid.top.html" 0 }
	{}
	{ "GNU Texinfo Manuals"
	    "http://www.cis.ohio-state.edu:85/info/dir" 0 }
	{ "X11R5 Man Pages"
	    "http://www.cis.ohio-state.edu:80/man/x11.top.html" 0 }
	{ "VMS Help System"
	    "http://slacvx.slac.stanford.edu:80/HELP" 0 }
	{ "Zippy the Pinhead"
	   "http://www.cis.ohio-state.edu:84/" 0}
    }}
    {{help "Help" 0 right}  {
       {"Overview" 	"tkWWWHelpGetTopic overview" 0}
       {"Index" 	"tkWWWHelpGetTopic index" 0}
       {"Keyboard" 	"tkWWWHelpGetTopic keyboard" 0}
       {"Tutorial" 	"tkWWWHelpGetTopic tutorial" 0}
       {}
       {"Send mail to developer..." "DLG:show . .send_mail_dialog" 0}
       {"Subscribe to mailing list tk-www@athena.mit.edu" \
	   "tkWWWHelpAddToMailingList" }
       {"Unsubscribe to mailing list tk-www@athena.mit.edu" \
           "tkWWWHelpRemoveFromMailingList" }
       {}
       {"Using Help" 	 "tkWWWHelpGetTopic using_help" 6}
       {"Product Information" "tkWWWFileAboutTkWWW" 0}
   }}}

set tkWWWConfigButtons {
    {home "Home" {tkWWWNavigateRecordAndGoto $tkWWWHomePage}}
    {back "Back" {tkWWWNavigateBacktrack}}
    {find "Find" {DLG:show . .find_dialog}}
    {goto "Goto..." {DLG:show . .goto_dialog}}
    {reload "Reload" tkWWWNavigateReload}
    {generate "Generate Source" {tkWWWEditGenerate}}
    {clone "Clone" {tkWWWNavigateClone [tkWWWNavigateGetAddress]}}
    {close_window "Close Window" "tkWWWFileQueryCloseWindow"}
    {exit "Exit" "tkWWWFileQueryExit"}
}

set tkWWWConfigToggles {
    {is_index "Is Index"}
    {has_annotation "Has Annotations"}
}

proc tkWWWConfigDisplay {} {
    global tkWWWHistoryList
    global tkWWWBookmarksList
    if [llength $tkWWWHistoryList] {
	.buttons.back configure -state normal
	.menu.history.m entryconfigure 0 -state normal
	.menu.history.m entryconfigure 1 -state normal
    } {
	.buttons.back configure -state disabled
	.menu.history.m entryconfigure 0 -state disabled
	.menu.history.m entryconfigure 1 -state disabled
    }
    if [tkWWWNavigateIsIndex] {
	.buttons.find configure -state normal
	.menu.file.m entryconfigure 7 -state normal
	.toggles.is_index select
	
    } {
	.buttons.find configure -state disabled
	.menu.file.m entryconfigure 7 -state disabled
	.toggles.is_index deselect
    }
    if [tkWWWAnnotateDisplay] {
	.toggles.has_annotation select
 	.menu.annotate.m entryconfigure 3 -state normal
    } {
	.toggles.has_annotation deselect
 	.menu.annotate.m entryconfigure 3 -state disabled
    }
    if [llength $tkWWWBookmarksList] {
	.menu.bookmarks.m entryconfigure 0 -state normal
    } {
	.menu.bookmarks.m entryconfigure 0 -state disabled
    }
}

set tkWWWConfigFileSuffixes   {
    { ".mime" "www/mime"  1.0 }	
    { ".PS" 	"application/postscript"  0.8 }
    { ".eps" 	"application/postscript"  0.8 }
    { ".ai" 	"application/postscript"  0.5 }
    { ".ps" 	"application/postscript"  0.8 }
    { ".html" "text/html"  1.0 }		
    { ".c" 	"text/plain"  0.5 }
    { ".h" 	"text/plain"  0.5 }		
    { ".m" 	"text/plain"  0.5 } 
    { ".txt"  "text/plain"  0.5 }
    { ".rtf" 	"application/x-rtf"  1.0 }
    { ".snd"  "audio/basic"  1.0 }
    { ".au"   "audio/basic"  1.0 }
    { ".bin" 	"application/octet-stream"  1.0 }
    { ".dmp.Z"  "image/compressed-x11-dump" 1.0}
    { ".dmp"  "image/x11-dump" 1.0}
    { ".pbm.Z" "image/compressed-portable-bitmap" 1.0}
    { ".pbm" "image/portable-bitmap" 1.0}
    { ".ppm.Z" "image/compressed-portable-pixmap" 1.0}
    { ".ppm" "image/portable-pixmap" 1.0}
    { ".xbm" "image/x-bitmap" 1.0} 
    { ".xbm.Z" "image/compressed-xbm" 1.0}
    { ".gif.Z" "image/compressed-gif" 1.0}
    { ".gif"  "image/gif"  1.0 }
    { ".tiff" "image/x-tiff"  1.0 }
    { ".jpg"  "image/jpeg"  1.0 }
    { ".JPG"  "image/jpeg"  1.0 }
    { ".JPEG" "image/jpeg"  1.0 }
    { ".jpeg" "image/jpeg"  1.0 }
    { ".ras"  "image/cmu-raster" 1.0}
    { ".MPEG" "video/mpeg"  1.0 }
    { ".mpg" "video/mpeg"  1.0 }
    { ".MPG" "video/mpeg"  1.0 }
    { ".mpeg" "video/mpeg"  1.0 }
    { ".tex" "application/tex" 1.0}
    { ".latex" "application/latex" 1.0}
    { ".texi" "application/texinfo" 1.0}
    { ".dvi" "application/dvi" 1.0}
    { ".tar" "archive/tar" 1.0}
    { ".tar.Z"  "archive/compressed-tar" 1.0}
    { ".shar" "archive/shar" 1.0}
    { ".shar.Z" "archive/compressed-shar" 1.0}
    { ".Z" 	"application/x-compressed" 0.5 }
    { ".z"    "application/x-zipped" 0.5}
    { ".x11.Z" "image/compressed-x11-dump" 1.0}
    { ".x11" "image/x11-dump" 1.0}
}

proc tkWWWConfigDisplayImageFile {file_name} {
    global tkWWWConfigImageViewer
    global tkWWWConfigImageDirectView
    if {$tkWWWConfigImageDirectView} {
	tkWWWOutputClearBody
	place .f.graphics -in .f.msg -relx 0.0 \
	    -rely 0.0 -relwidth 0.0 -relheight 0.0
	update
	set output [open  "| $tkWWWConfigImageViewer \
-windowid [winfo id .f.graphics] -private -at 1,1 $file_name" "r"]
	while {[gets $output string] != -1} {
	    .f.msg insert insert "$string\n"
	    update
	} 
	close $output
	place .f.graphics -in .f.msg -relx 0.0 \
	    -rely 0.0 -relwidth 1.0 -relheight 1.0
	tkWWWOutputSetMessage {}
    } {
	tkWWWNavigatePreface {Image file
To display press the button below and wait a few seconds
} {Press here to show image} {Displaying image ...} \
"exec $tkWWWConfigImageViewer $file_name &"
    }
}

proc tkWWWConfigDisplayFile {type file_name} {
    case $type {
	"image/*" {
	    tkWWWConfigDisplayImageFile $file_name
	}
	"video/*" {
	global tkWWWConfigVideoViewer
	tkWWWNavigatePreface {Video file
To display press the button below and wait a few seconds
} {Press here to show video} {Displaying video ...} \
"exec $tkWWWConfigVideoViewer $file_name &"
	}
	"audio/*" {
	global tkWWWConfigAudioViewer
	tkWWWNavigatePreface {Audio file
To display press the button below and wait a few seconds
} {Press here to show audio} {Displaying audio ...} \
"exec $tkWWWConfigAudioViewer $file_name &"
	}
	{application/postscript} {
	global tkWWWConfigPostscriptViewer
	tkWWWNavigatePreface {Postscript file
To display press the button below and wait a few seconds
} {Press here to show postscript} {Displaying Postscript ...} \
"exec $tkWWWConfigPostscriptViewer $file_name &"
	}
	{application/dvi} {
	global tkWWWConfigDviViewer
	tkWWWNavigatePreface {TeX Device Independent file
To display press the button below and wait a few seconds
} {Press here to show DVI file} {Displaying TeX DVI file ...} \
"exec $tkWWWConfigDviViewer $file_name &"
	}
	{text/plain} {
	    tkWWWOutputSetBodyToFile $file_name
	} 
	{default} {
	    case [DLG:question . .graphics_dialog \
		  "Unknown file type" \
		      "Display as text" "Display as image" "Save file as..."] {
			  {1} {tkWWWOutputSetBodyToFile $file_name}
			  {2} {tkWWWConfigDisplayImageFile $file_name}
			  {3} {tkWWWOutputSaveFileAs $file_name}
		      }
	}
    }
}




/**/## *************************
/**/## Set options
/**/## *************************

option add *background gray userDefault
option add *foreground black userDefault
option add *buttons*relief raised userDefault
option add *buttons*borderWidth 2 userDefault
option add *toggles*relief flat userDefault
option add *toggles*borderWidth 2 userDefault
option add *Scrollbar.foreground gray userDefault
option add *Hypertext.Button.relief raised userDefault
option add *Frame.relief raised userDefault
option add *Frame.borderWidth 2 userDefault
option add *Hypertext*background bisque userDefault
option add *Text*background bisque userDefault
option add *Text*wrap word userDefault
option add *Text.relief sunken userDefault
option add *Text.borderWidth 2 userDefault
option add *Scrollbar*relief sunken userDefault

/**/## *************************
/**/## Set key bindings
/**/## *************************

proc GetSelection {} {# xf ignore me 7
/**/##########
/**/# Procedure: GetSelection
/**/# Description: get current selection
/**/# Arguments: none
/**/# Returns: none
/**/# Sideeffects: none
/**/##########

/**/# the save way
  set xfSelection ""
  catch "selection get" xfSelection
  if {"$xfSelection" == "selection doesn't exist or form \"STRING\" not defined"} {
    return ""
  } {
    return $xfSelection
  }
}

bind Text <Left> {# xf ignore me 9
  %W mark set insert insert-1c; %W yview -pickplace insert}
bind Text <Right> {# xf ignore me 9
  %W mark set insert insert+1c; %W yview -pickplace insert}
bind Text <Up> {# xf ignore me 9
  %W mark set insert insert-1l; %W yview -pickplace insert}
bind Text <Down> {# xf ignore me 9
  %W mark set insert insert+1l; %W yview -pickplace insert}
bind Text <Delete> {# xf ignore me 9
  tk_textBackspace %W; %W yview -pickplace insert}
bind Text <Control-a> {# xf ignore me 9
  %W mark set insert "insert linestart"; %W yview -pickplace insert}
bind Text <Control-b> {# xf ignore me 9
  %W mark set insert insert-1c; %W yview -pickplace insert}
bind Text <Control-d> {# xf ignore me 9
  %W delete insert insert+1c; %W yview -pickplace insert}
bind Text <Control-e> {# xf ignore me 9
  %W mark set insert "insert lineend"; %W yview -pickplace insert}
bind Text <Control-f> {# xf ignore me 9
  %W mark set insert insert+1c; %W yview -pickplace insert}
bind Text <Meta-h> {# xf ignore me 9
  %W delete "insert linestart" insert; %W yview -pickplace insert}
bind Text <Control-i> {# xf ignore me 9
  %W inser insert \t; %W yview -pickplace insert}
bind Text <Control-j> {# xf ignore me 9
  %W inser insert \n; %W yview -pickplace insert}
bind Text <Control-k> {# xf ignore me 9
  if {[%W compare insert == "insert lineend"]} {
    %W delete insert "insert lineend +1c"
    %W yview -pickplace insert
  } {
    %W delete insert "insert lineend"
    %W yview -pickplace insert}}
bind Text <Control-l> {# xf ignore me 9
  %W yview -pickplace insert}
bind Text <Control-m> {# xf ignore me 9
  %W inser insert \n; %W yview -pickplace insert}
bind Text <Control-n> {# xf ignore me 9
  %W mark set insert insert+1l; %W yview -pickplace insert}
bind Text <Control-o> {# xf ignore me 9
  %W inser insert \n; %W yview -pickplace insert}
bind Text <Control-p> {# xf ignore me 9
  %W mark set insert insert-1l; %W yview -pickplace insert}
bind Text <Meta-v> {# xf ignore me 9
  set xfCounter [lindex [%W config -height] 4]
  while {$xfCounter > 0} {
    %W mark set insert insert-1l
    incr xfCounter -1
  }
  %W yview -pickplace insert}
bind Text <Control-v> {# xf ignore me 9
  set xfCounter [lindex [%W config -height] 4]
  while {$xfCounter > 0} {
    %W mark set insert insert+1l
    incr xfCounter -1
  }
  %W yview -pickplace insert}
bind Text <Control-w> {# xf ignore me 9
  catch "%W delete sel.first sel.last"}
bind Text <Control-y> {# xf ignore me 9
  %W insert insert "[GetSelection]"; %W yview -pickplace insert}
bind Text <Home> {# xf ignore me 9 
  %W mark set insert 1.0; %W yview -pickplace insert}
bind Text <End> {# xf ignore me 9
  %W mark set insert end; %W yview -pickplace insert}
bind Text <Prior> {# xf ignore me 9
  set xfCounter [lindex [%W config -height] 4]
  while {$xfCounter > 0} {
    %W mark set insert insert-1l
    incr xfCounter -1
  }
  %W yview -pickplace insert}
bind Text <Next> {# xf ignore me 9
  set xfCounter [lindex [%W config -height] 4]
  while {$xfCounter > 0} {
    %W mark set insert insert+1l
    incr xfCounter -1
  }
  %W yview -pickplace insert}


if {$tkVersion >= 3.0} {
bind Entry <Right> {# xf ignore me 9
  %W icursor [expr [%W index insert]+1]; tk_entrySeeCaret %W }
bind Entry <Left> {# xf ignore me 9
  %W icursor [expr [%W index insert]-1]; tk_entrySeeCaret %W }
bind Entry <Control-a> {# xf ignore me 9
  %W icursor 0; %W view 0}
bind Entry <Control-b> {# xf ignore me 9
  %W icursor [expr [%W index insert]-1]; tk_entrySeeCaret %W}
bind Entry <Control-d> {# xf ignore me 9
  %W delete [%W index insert]; tk_entrySeeCaret %W }
bind Entry <Control-e> {# xf ignore me 9
  %W icursor end; tk_entrySeeCaret %W}
bind Entry <Control-f> {# xf ignore me 9
  %W icursor [expr [%W index insert]+1]; tk_entrySeeCaret %W}
bind Entry <Control-h> {# xf ignore me 9
  tk_entryBackspace %W; tk_entrySeeCaret %W}
bind Entry <Meta-h> {# xf ignore me 9
  %W delete 0 insert; %W view insert}
bind Entry <Control-i> {# xf ignore me 9
  %W insert insert \t; tk_entrySeeCaret %W}
bind Entry <Control-k> {# xf ignore me 9
  %W delete insert end; tk_entrySeeCaret %W}
bind Entry <Control-l> {# xf ignore me 9
  tk_entrySeeCaret %W}
bind Entry <Control-v> {# xf ignore me 9
  %W insert insert "[GetSelection]"; tk_entrySeeCaret %W}
bind Entry <Control-w> {# xf ignore me 9
  catch "%W delete sel.first sel.last"; tk_entrySeeCaret %W}
bind Entry <Control-y> {# xf ignore me 9
  %W insert insert "[GetSelection]"; tk_entrySeeCaret %W}
} {
bind Entry <Right> {# xf ignore me 9
  %W cursor [expr [%W index cursor]+1]; tk_entrySeeCaret %W }
bind Entry <Left> {# xf ignore me 9
  %W cursor [expr [%W index cursor]-1]; tk_entrySeeCaret %W }
bind Entry <Control-a> {# xf ignore me 9
  %W cursor 0; %W view 0}
bind Entry <Control-b> {# xf ignore me 9
  %W cursor [expr [%W index cursor]-1]; tk_entrySeeCaret %W}
bind Entry <Control-d> {# xf ignore me 9
  %W delete [%W index cursor]; tk_entrySeeCaret %W }
bind Entry <Control-e> {# xf ignore me 9
  %W cursor end; tk_entrySeeCaret %W}
bind Entry <Control-f> {# xf ignore me 9
  %W cursor [expr [%W index cursor]+1]; tk_entrySeeCaret %W}
bind Entry <Control-h> {# xf ignore me 9
  tk_entryBackspace %W; tk_entrySeeCaret %W}
bind Entry <Meta-h> {# xf ignore me 9
  %W delete 0 cursor; %W view cursor}
bind Entry <Control-i> {# xf ignore me 9
  %W insert cursor \t; tk_entrySeeCaret %W}
bind Entry <Control-k> {# xf ignore me 9
  %W delete cursor end; tk_entrySeeCaret %W}
bind Entry <Control-l> {# xf ignore me 9
  tk_entrySeeCaret %W}
bind Entry <Control-v> {# xf ignore me 9
  %W insert cursor "[GetSelection]"; tk_entrySeeCaret %W}
bind Entry <Control-w> {# xf ignore me 9
  catch "%W delete sel.first sel.last"; tk_entrySeeCaret %W}
bind Entry <Control-y> {# xf ignore me 9
  %W insert cursor "[GetSelection]"; tk_entrySeeCaret %W}
}
bind Entry <2> {# xf ignore me 9
  %W insert cursor "[GetSelection]"; tk_entrySeeCaret %W}
