## edit.tcl main file for tkWWW  user interface
## ==============
## Copyright (C) 1992 Joseph Wang
## See the file COPYRIGHT for conditions

## The procedures in this file are responsible for editing inside of tkWWW.

## Conventions:
##   All procedures in this file begin with tkWWWEdit

## Bugs and Todo:
##   The UI sucks rocks.  Need to get people to use it and comment on
##      problems
##   Some abstraction questions

# This is the style that the editor is current writing in
set tkWWWEditCurrentStyle "Normal"

# This is the cut buffer
set tkWWWEditSelection {}

## tkWWWEditInitialize:
## ----------
## Initialize editing commands
##    Generate the edit dialog
##    Bind editing commands onto the display text widget

proc tkWWWEditInitialize {} {
   DLG:toplevel . .edit

   set button_list {
       {close "Close" "tkWWWEditClose"}
       {generate "Generate" "tkWWWEditGenerate"}
       {save "Save" "tkWWWEditSave"}
   }

   pack append .edit  \
      [ tkWWWOutputMakeButtons .edit.button_box $button_list ] {top fillx} \
      [ frame .edit.f -relief sunken -borderwidth 2] {bottom expand fill}

   pack append .edit.f \
      [ text .edit.f.text \
          -wrap none \
          -yscrollcommand {.edit.vs set } ] { expand fill padx 2 pady 2}

   pack before .edit.f \
    [scrollbar .edit.vs -command { .edit.f.text yview } \
           -orient vertical] { right filly }

   tkWWWEditBindText .f.msg
}

proc tkWWWEditBindText {w} {
    bind Text <Any-KeyPress> {
	tkWWWEditInsertCharacter %W %A
    }
}

proc tkWWWEditInsertCharacter {w a} {
    global tkWWWEditCurrentStyle
    if {"$a" != ""} {
	set start [$w index insert]
	$w insert insert $a
	$w yview -pickplace insert

	foreach tag [$w tag names $start] {
	    $w tag remove $tag $start insert
	}
	$w tag add $tkWWWEditCurrentStyle $start insert
    }
}

## ********************
## Procedures to show or hide the edit window
## ********************

proc tkWWWEditOpen {} {
    .edit.f.text delete 1.0 end
    DLG:show . .edit
}

proc tkWWWEditClose {} {
    DLG:hide .edit
}

## ********************
## Save Generated HTML to a file
## ********************

proc tkWWWEditSaveHTML {} {
    global fsBox
    set address [tkWWWNavigateGetAddress]
    set list [split $address ":"]
    set type [lindex $list 0]
    set name [lindex $list 1]
    set file "file.html"
    set fsBox(path) ""

    if {"$type" == "file" && ![string match $name "//*"]} {
	set fsBox(path) [file dirname $name]
	set file [file tail $name]
    }
    tkWWWEditClose
    FSBox {Save HTML} $file \
	{tkWWWEditOutputHTML $fsBox(path)/$fsBox(name)}
    DLG:show . .fsBox
}

proc tkWWWEditOutputHTML {filename} {
    set file [open $filename "w"]
    puts $file [.edit.f.text get 0.0 end]
    close $file
}

## ********************
## Procedures to generate HTML in the edit window from 
## stuff displayed in the display window
## ********************

proc tkWWWEditGenerate {} {
    tkWWWEditOpen
    tkWWWEditCopyText .f.msg .edit.f.text
    tkWWWEditCopyTags .f.msg .edit.f.text
    tkWWWEditConvertToHTML .edit.f.text
}


proc tkWWWEditCopyText {from to} {
    $to delete 1.0 end
    $to insert 1.0 [$from get 1.0 end]
}

proc tkWWWEditCopyTags {from to} {
    foreach item [$from tag names] {
	set range [$from tag ranges $item] 
	set length [llength $range]
	for {set i 0} {$i < $length} {incr i 2} {
	    set j $i
	    incr j 1
	    $to tag add $item [lindex $range $i] [lindex $range $j] 
	}
    }
}

proc tkWWWEditConvertToHTML {w} {
    $w insert 1.0 "<TITLE>[tkWWWNavigateGetTitle]</TITLE>\n"

    tkWWWEditDeleteTextWithTag $w offset
    tkWWWEditDeleteTextWithTag $w blankline
    tkWWWEditDeleteTextWithTag $w softbreak

    tkWWWEditBreakLines $w
    
    for {set n 1} {$n < 7} {incr n} {
	tkWWWEditDelimitTag $w Heading$n "<H$n>" "</H$n>\n"
    }
    tkWWWEditDelimitTag $w Address <ADDRESS> </ADDRESS>
    tkWWWEditDelimitTag $w Preformatted "<PRE>\n" "</PRE>\n"
    tkWWWEditDelimitTag $w Menu "<MENU>\n" "</MENU>\n"
    tkWWWEditDelimitTag $w Glossary "<DL>\n" "</DL>\n"
    tkWWWEditDelimitTag $w List "<UL>\n" "</UL>\n"
    tkWWWEditDelimitTag $w Example "<XMP>\n" "</XMP>\n"
    tkWWWEditDelimitTag $w Dir "<DIR>\n" "</DIR>\n"

    tkWWWEditConvertBullets $w ListItem <LI>
    tkWWWEditConvertBullets $w GlossaryDefinition <DD>
    tkWWWEditConvertBullets $w GlossaryTerm <DT>
    tkWWWEditConvertBullets $w ParagraphBreak "<P>\n"
    tkWWWEditConvertAnchors $w
}

proc tkWWWEditConvertAnchors {w} {
    set i 1.0
    while {[set range [$w tag nextrange "Anchor" $i]] != ""} {
	$w mark set tmp [lindex $range 1]
	foreach t [$w tag names [lindex $range 0]] {
	    if [regexp {h([0-9]+)} $t {} number] {
		set anchor \
		    [tkWWWServerSend [list HtGetReference $number]]
		if {[lindex $anchor 1] != ""} {
		    $w insert [lindex $range 0] \
			"<A NAME=\"[lindex $anchor 0]\" HREF=\"[lindex $anchor 1]\">"   
		} {
		    $w insert [lindex $range 0] "<A NAME=\"[lindex $anchor 0]\">"		} 
		$w insert tmp "</A>"
	    }
	}
	set i [$w index tmp]
    }
}

proc tkWWWEditBreakLines {w} {
    for {set i 1} {[$w compare $i.0 < end]} {incr i} {
	set chars [lindex [split [$w index "$i.0 lineend"] "."] 1]
	if {$chars > 80 && [$w compare "$i.80 wordstart" > "$i.0"]} {
	    $w mark set tmp "$i.80 wordstart"
	    if {[$w get tmp] == " "} {
		$w delete tmp
	    } {
		$w delete "tmp - 1 char"
	    }
	    $w insert tmp "\n"
	}
    }
}

proc tkWWWEditConvertBullets {w text_tag sgml_tag} {
    $w mark set tmp 1.0
    while {[set range [$w tag nextrange $text_tag tmp]] != ""} {
	$w mark set tmp [lindex $range 0]
	$w delete tmp [lindex $range 1]
	$w mark set insert [$w index tmp]
	tkWWWEditInsertClean $w $sgml_tag
    }
}

proc tkWWWEditDeleteTextWithTag {w tag} {
    $w mark set tmp 1.0
    while {[set range [$w tag nextrange $tag tmp]] != ""} {
	$w mark set tmp [lindex $range 0]
	$w delete tmp [lindex $range 1]
    }
}

proc tkWWWEditDelimitTag {w tag begin end} {
    $w mark set tmp 1.0
    while {[set range [$w tag nextrange $tag tmp]] != ""} {
	$w mark set tmp [lindex $range 1]
	
	$w mark set insert [lindex $range 0]
	tkWWWEditInsertClean $w $begin

	$w mark set insert [$w index tmp]
	tkWWWEditInsertClean $w $end
    }
}

proc tkWWWEditInsertClean {w text} {
    set start [$w index insert]
    $w insert insert $text
    foreach tag [$w tag names $start] {
	$w tag remove $tag $start insert
    }
}

## ********************
## These procedure are used to edit the display window
## ********************


proc tkWWWEditQueryChangeTitle {w} {
    DLG:entry . .change_title "Change Title Dialog" \
	{"Title:"} {Change title} "tkWWWEditChangeTitle"
    DLG:show . .change_title
}    


proc tkWWWEditChangeTitle {title} {
    tkWWWOutputSetAddress [tkWWWNavigateGetAddress] $title
}

## tkWWWEditAddAnchor
## -------------

## Turn currently highlighted area into an anchor

proc tkWWWEditBeginAnchor {w} {
    if {[.f.msg tag ranges sel] != {}} {
	DLG:entry . .anchor_dialog "Anchor Dialog" \
	    {"Name:" "Href:"} {Begin} "tkWWWEditAddAnchor"
	DLG:show . .anchor_dialog
    }
}
 
proc tkWWWEditAddAnchor {name href} {
    set number [tkWWWServerSend [list "HtAddAnchor" $name $href]]

    tkW3HtBindAnc $number
    if {[.f.msg tag ranges sel] != {}} {
	set start [.f.msg index sel.first]
	set finish [.f.msg index sel.last]
	foreach tag [.f.msg tag names $start] {
	    if [regexp {h(0-9)+} $tag] {
		.f.msg tag remove $tag $start $finish
	    }
	}
	.f.msg tag add Anchor $start $finish 
	.f.msg tag add h$number $start $finish
	.f.msg tag remove sel $start $finish
    }
}

proc tkWWWEditTagSelected {w tag} {
    global tkWWWEditCurrentStyle
    set list [tkWWWEditClearTagsFromSelected $w]
    if {$list != {}} {
	$w tag add $tag [lindex $list 0] [lindex $list 1]
	$w mark set insert [lindex $list 1]
    }
    set tkWWWEditCurrentStyle $tag
}

proc tkWWWEditBeginListItem {w} {
    global tkWWWEditCurrentStyle
    tkW3HtBuf List "\n"
    set start [$w index insert]
    tkW3HtInsert "* " ListItem
    set tkWWWEditCurrentStyle List
    $w tag add $tkWWWEditCurrentStyle $start insert
}
    
proc tkWWWEditClearTagsFromSelected {w} {
    global tkWWWEditCurrentStyle

    if {[$w tag ranges sel] != {}} {
	set start [$w index sel.first]
	set finish [$w index sel.last]
	foreach tag [.f.msg tag names $start] {
	    $w tag remove $tag $start $finish
	}
	return [list $start $finish]
    }
    set tkWWWEditCurrentStyle Normal
    return {}
}

proc tkWWWEditCut {w} {
    tkWWWEditCopy $w
    tkWWWEditDelete $w
}

proc tkWWWEditCopy {w} {
    global tkWWWEditSelection
    set tkWWWEditSelection [$w get sel.first sel.last]
}

proc tkWWWEditDelete {w} {
    $w delete sel.first sel.last
}

proc tkWWWEditPaste {w} {
    global tkWWWEditSelection
    if {"" != [$w tag ranges sel]} {
	$w delete sel.first sel.last
	$w set mark insert sel.first
    }
    $w insert insert "$tkWWWEditSelection"
}

proc tkWWWEditClear {w} {
    $w delete 1.0 end
}

proc tkWWWEditInsertBullet {w type} {
    global tkWWWEditCurrentStyle
    case $type {
	"LI" {
	    tkW3HtBuf offset " "
	    tkW3HtInsert "* " ListItem
	}
	"DD" {
	    tkW3HtInsert "* " GlossaryDefinition
	}
	"DT" {
	    tkW3HtInsert "* " GlossaryTerm
	}
	"P" {
	    tkW3HtInsert "\n\n" ParagraphBreak
	}
    }
}

proc tkWWWEditSelectAll {w} {
    $w tag add sel 1.0 end
}

proc tkWWWEditDeselectAll {w} {
    $w tag remove sel 1.0 end
}

proc tkWWWEditCreateNewPage {page_name} {
    tkWWWOutputSetAddress $page_name ""
    tkWWWEditQueryChangeTitle .f.msg
    tkWWWOutputClearBody
}
