## hypertext.tcl Procedures for processing the hypertext
## ==============
## Copyright (C) 1992 Joseph Wang
## See the file COPYRIGHT for conditions

## Conventions:
## To reduce traffic between server and client, the procedures in this
## procedure have abbreviated file names which begin with tkW3Ht

# Use the standard tk text widget to display text
# Begin an environment

# The procedure below inserts text into a given text widget and
# applies one or more tags to that text.  The arguments are:

# text          Text to insert (it's inserted at the "insert" mark)
# args          One or more tags to apply to text.  If this is empty
#               then all tags are removed from the text.

proc tkW3HtInsert {text args} {
    set start [.f.msg index insert]
    .f.msg insert insert $text
    foreach tag [.f.msg tag names $start] {
        .f.msg tag remove $tag $start insert
    }
    foreach tag $args {
	.f.msg tag add $tag $start insert
    }
}

# This exists so that the button events are only bound once
set tkW3HtMaxAnchorNumber 0

proc tkW3HtBeginDoc {} {
    .f.msg configure -state normal
    .f.msg delete 1.0 end
}
    
# This returns a string describing the contents of an anchor
proc tkW3HtGetLinkInfo {link_number} {
    set return_string ""
    set name [tkWWWNavigateGetAnchorName $link_number]
    set dest [tkWWWNavigateGetAnchorDestination $link_number]
    append return_string "Name: $name\n"
    if {$dest != ""} {
	append return_string "Destination: $dest"
    } {
	append return_string "Internal anchor"
    }
    return $return_string
}


# Add a string to the tkW3 buffer
proc tkW3HtBuf {style text} {
    set start [.f.msg index insert]
    .f.msg insert insert $text
    foreach tag [.f.msg tag names $start] {
        .f.msg tag remove $tag $start insert
    }
    .f.msg tag add $style $start insert
} 

proc tkW3HtAnc {n style w} {
    set list [split $w "\n"]
    set head [lindex $list 0]
    set tail [lrange $list 1 end]
    tkW3HtInsert  $head h$n Anchor $style
    foreach item $tail {
	tkW3HtInsert "\n" softbreak
	tkW3HtInsert $item h$n Anchor $style
    }
} 


proc tkW3HtEndDoc {} {
}

proc tkW3HtBindAnc {n} {
    global tkW3HtMaxAnchorNumber
    for {set i $tkW3HtMaxAnchorNumber} {$i <= $n} {incr i 1} {
	.f.msg tag bind h$i <1> "tkWWWNavigateSelectReference $i"
	.f.msg tag bind h$i <2> "tkWWWNavigateCloneReference $i"
	.f.msg tag bind h$i <3> "DLG:information . .info_dialog \"Anchor
\[tkW3HtGetLinkInfo $i\]\" {OK}"
    }
    set tkW3HtMaxAnchorNumber $n
}


 # Blanklines
proc tkW3HtBlankLines {nlines args} {
    for {set i 0} {$i < $nlines} {incr i} {
	tkW3HtInsert "\n" blankline
    }
}


proc tkW3HtBul {style type} {
    case $type {
	"LI" {
	    tkW3HtInsert "* " $style ListItem
	}
	"DD" {
	    tkW3HtInsert "* " $style GlossaryDefinition
	}
	"DT" {
	    tkW3HtInsert "* " $style GlossaryTerm
	}
	"P" {
	    tkW3HtInsert "\n" $style ParagraphBreak
	}
    }
}
