## navigate.tcl Procedures for navigation
## ==============
## Copyright (C) 1992 Joseph Wang
## See the file COPYRIGHT for conditions

## The procedures in this file are responsible for navigation between 
## different hypertext pages on World Wide Web.  They are called by the
## objects in other tkWWW files in response to a user's request to 
## display a WWW page.

## Conventions:
##   All procedures in this file begin with tkWWWNavigate

proc tkWWWNavigateBacktrack {} {
    if {[set backpage [tkWWWHistoryPop]] !={}} {
	tkWWWNavigateGoto [lindex $backpage 1] [lindex $backpage 3]
    } {
	tkWWWOutputError "Error: No pages previous to this"
    }
}

proc tkWWWNavigateClone {{page ""}} {
  global tkWWWSourcePath
  global tkWWWHomePage
  
  if {$page == ""} {
      set page [tkWWWNavigateGetAddress]
  }

  tkWWWOutputSetMessage "Opening New Window ..." 5000
  exec tkWWW -wwwstart [tkWWWServerSend [list HtParseName $page]] \
            -wwwhome $tkWWWHomePage -file $tkWWWSourcePath/main.tcl \
            -wwwnoinfo -name tkWWW &
}

# This closes this tkWWW window
proc tkWWWNavigateCloseThisWindow {} {
  tkWWWServerClose
# Necessary to clear name from interpreter table
  destroy .
  exit
}

# This closes all tkWWW windows 
proc tkWWWNavigateCloseAllWindows {} {
    set self [winfo name .]
    foreach interp [winfo interps] {
	if {[regexp {tkWWW.*} $interp] && "$interp" != "$self"} {
	    catch "send \"$interp\" {after 1 {tkWWWNavigateCloseThisWindow}}" 
	}
    }
    tkWWWNavigateCloseThisWindow
}
	


## tkWWWNavigateFind: Does keyword search off current page
## ----------
## Arguments:
##   key - a space separated list of keywords
## No return value
##
## This procedure performs a keyword search off the currently displayed
## page.  It works by converting the keywords and current page into a
## World Wide Web address, and then calls the procedure tkWWWNavigateGoto
## and goes to that page.
##

proc tkWWWNavigateFind {key} {
    regsub -all {[ ]+} $key {+} key
    set address [string trim [tkWWWNavigateGetAddress] " \n?"]
    set address [lindex [split $address "?"] 0]
    append address "?" $key
    tkWWWNavigateRecordAndGoto $address
}

# Some abstraction barriers
# Note that tcl returns the value of the last statement executed

proc tkWWWNavigateGetAddress {} {
   .titles.address_entry get
}

proc tkWWWNavigateGetTitle {} {
   .titles.title_entry get
}

proc tkWWWNavigateGetSource {} {
   tkWWWServerSend HtGetSource
}

proc tkWWWNavigateGoto {page {scroll_position ""}} {
    # Change the cursor to a busy signal and tell user we are getting file
    tkWWWOutputCursorWait
    tkWWWOutputSetMessage "Getting page ..."
    catch {place forget .f.graphics}
    update idletask
    
    if {![string match "#*" $page]} {
	# case on the type of page
	# wish doesn't like comments in case statements
	case [set page_name [tkWWWServerSend [list HtParseName $page]]] {
	    {telnet:* tn3270:* rlogin:*} {
		tkWWWTelnetPage $page_name
	    }
	    {exec:*} {
		# Find the command
		if [regexp {^exec:(.*)} $page_name {} command] {
		    regsub -all %20 $command " " command

		    tkWWWOutputSetAddress $page_name {}
		    
		    tkWWWNavigatePreface "System Call
Execute system command \"$command\"
" {Push here to execute System Call} "Executing System Call \"$command\"..." \
"exec $command &"
		}
	    }
	    default {
		# if the file type if file, http, gopher, or wais we need 
		# to load the page in from the server
		tkWWWNavigateLoadPageFromServer $page_name
	    }
	} 
	# Set the output to the correct line
	if {$scroll_position != ""} {
	    tkWWWOutputSetScrollPosition $scroll_position
	}
    } {
	set page_name $page
    }

    # Goto to the anchor
    set page_tag ""
    regexp {#(.*)$} $page_name {} page_tag
    if {$page_tag != ""} {
	set number [tkWWWServerSend [list HtGetNumberFromTag $page_tag]]
	if {$number != ""} {
	    .f.msg yview -pickplace h$number.first
	}
    }

    # Display
    tkWWWConfigDisplay 

    # Clear the message and set the cursor back to normal
    tkWWWOutputSetMessage {}
    tkWWWOutputCursorNormal
}

proc tkWWWNavigateLoadPageFromServer {page_name} {
    # Chop off the anchor
    regexp {^[^#]*} $page_name page_value

    # Load the page and get the return code
    # If the return code contains "Error" signal an error
    if [regexp {Error:.*} [tkWWWServerSend [list HtLoad $page_value]] error] {
	# Ooops there's an error
	# If it is a "Page Not Found" error create new page
	if [regexp {Page not found.*file:.*} $error] {
	    if {[DLG:question . .page_not_found "$error
Do you wish to create?" "Yes" "No"]==1} {
		     tkWWWEditCreateNewPage $page_name
		 } {
		     tkWWWHistoryRestore
		 }
	    return
	}
	# Undo any changes we have made to the history list
	tkWWWHistoryRestore
	# Tell the user about the error
	tkWWWOutputError $error
    } { 
	# No error from backend, flip to new page
        # What we should get back from the backend are
        # commands to generate the hypertext.
        tkWWWOutputSetMessage "Formatting page ..."
	update idletasks
	eval [tkWWWServerSend HtGetText]
	eval tkWWWOutputSetAddress [tkWWWServerSend HtGetAddressAndTitle]
    }
}

proc tkWWWNavigatePreface {body_text button_text message command} {
    tkWWWOutputClearBody
    tkW3HtInsert "$body_text"
    tkW3HtInsert "$button_text" action Anchor
    .f.msg tag bind action <1> "tkWWWOutputSetMessage {$message} 5000
$command"
}
    
proc tkWWWNavigateSaveSource {filename} {
   set file [open $filename w]
   puts $file [tkWWWServerSend HtGetSource]
   close $file
} 

proc tkWWWNavigateSaveText {filename} {
   set file [open $filename a+]
   puts $file [.f.msg get 0.0 end]
   close $file
} 

proc tkWWWNavigatePrintUsingCommand {command} {
    set file [open "| $command" w+]
    puts $file [.f.msg get 0.0 end]
    close $file
}

proc tkWWWNavigateMailText {to subject} {
    tkWWWConfigSendMail $to $subject [.f.msg get 0.0 end]
}

proc tkWWWNavigateSelectReference {n} {
    tkWWWNavigateRecordAndGoto [tkWWWNavigateGetAnchorDestination $n]
}

proc tkWWWNavigateGetAnchorName {n} {
    lindex [tkWWWServerSend [list HtGetReference $n]] 0
}

proc tkWWWNavigateGetAnchorDestination {n} {
    lindex [tkWWWServerSend [list HtGetReference $n]] 1
}

proc tkWWWNavigateCloneReference {n} {
    tkWWWNavigateClone [tkWWWNavigateGetAnchorDestination $n]
}

# Record current address on the history list and go to the requested page

proc tkWWWNavigateRecordAndGoto {page {scroll ""}} {
    tkWWWHistoryRecord \
	[tkWWWNavigateGetTitle] [tkWWWNavigateGetAddress] \
	[tkWWWOutputGetScrollPosition]
    tkWWWNavigateGoto $page
}

proc tkWWWNavigateDelayedGoto {page {scroll ""}} {
    after 1 [list tkWWWNavigateGoto $page $scroll]
}

proc tkWWWNavigateUncache {address} {
    tkWWWServerSend [list HtUncache $address]
}

proc tkWWWNavigateReload {} {
    set address [tkWWWNavigateGetAddress]
    tkWWWNavigateUncache $address
    tkWWWNavigateGoto $address
}

proc tkWWWNavigateIsIndex {} {
    tkWWWServerSend "HtGetIsIndex"
}

proc tkWWWNavigateInitialize {} {
    global tkWWWConfigFileSuffixes

    foreach i $tkWWWConfigFileSuffixes {
	tkWWWServerSend [concat HtSetSuffix $i]
    }
}
