## output.tcl: output for tkWWW  user interface
## ==============
## Copyright (C) 1992 Joseph Wang
## See the file COPYRIGHT for conditions

# All output should go through these procedures

proc tkWWWOutputSetMessage {message {duration ""}} {
    .message configure -state normal
    .message delete 0 end
    .message insert 0 $message
    .message configure -state disabled
    if {$duration != ""} {
	after $duration {tkWWWOutputSetMessage {}}
    }
}

proc tkWWWOutputSetAddress {address title} {
    .titles.address_entry configure -state normal
    .titles.title_entry configure -state normal

    .titles.address_entry delete 0 end
    .titles.title_entry delete 0 end

    .titles.address_entry insert 0 $address
    .titles.title_entry insert 0 $title

    .titles.address_entry configure -state disabled
    .titles.title_entry configure -state disabled
}

proc tkWWWOutputSetBodyToFile {filename} {
    .f.msg delete 0.0 end
    set file [open $filename "r"]
    while {[gets $file line] != -1} {
	.f.msg insert insert "$line\n"
    }
    close $file
    .f.msg tag add Preformatted 0.0 end
}

proc tkWWWOutputGetScrollPosition {} {
    lindex [.vs get] 2
}

proc tkWWWOutputSetScrollPosition {value} {
    .f.msg yview $value
}

## ************************
## Cursor procedures
## ************************

proc tkWWWOutputCursorWait {} {
  .f.msg configure -cursor watch
}

proc tkWWWOutputCursorNormal {} {
  .f.msg configure -cursor xterm
}

proc tkWWWOutputMakeButtons {w list} {
    if [winfo exists $w] {
	foreach child [winfo children $w] {
	    destroy $child
	}
    } {
	frame $w -borderwidth 0
    }
    foreach item $list {
	pack append $w \
	    [frame $w.[lindex $item 0 ]_frame -borderwidth 3 \
	     -relief flat ] {left }
	pack append $w.[lindex $item 0]_frame \
	    [button $w.[lindex $item 0] \
	     -text [lindex $item 1] -command [lindex $item 2]] {left fill}
    }
    return $w
}

proc tkWWWOutputDoNothing {} {
}

proc tkWWWOutputMakeToggles {w list} {
    if [winfo exists $w] {
	foreach child [winfo children $w] {
	    destroy $child
	}
    } {
	frame $w -borderwidth 0
    }
    foreach item $list {
	set name [lindex $item 0]
	pack append $w \
	    [ checkbutton $w.$name -text [lindex $item 1] -variable $name] right
	bind $w.$name <Any-Enter> "tkWWWOutputDoNothing"
	bind $w.$name <Any-Leave> "tkWWWOutputDoNothing"
	bind $w.$name <1> "tkWWWOutputDoNothing"
	bind $w.$name <ButtonRelease-1> "tkWWWOutputDoNothing"
    }
    pack append $w \
	[label .message_label -text "Message:"] {left} \
	[entry .message -state disabled] {left  expand fillx}
    return $w
}

proc tkWWWOutputMakeTitleBox {w} {
    if [winfo exists $w] {
	foreach child [winfo children $w] {
	    destroy $child
	}
    } {
	frame $w -borderwidth 0
    }

    pack append $w \
	[frame $w.address_frame -borderwidth 0] {top fillx expand} \
	[frame $w.title_frame -borderwidth 0] {bottom fillx expand}

    pack append $w.address_frame \
	[label $w.address_label -text "Document Address:"] {left} \
	[entry $w.address_entry -state disabled ] {right fillx expand}

    pack append $w.title_frame \
	[label $w.title_label -text "Document Title:"] {left} \
	[entry $w.title_entry -state disabled ] {right fillx expand}

    return $w
}

proc tkWWWOutputError {message {fatal 0}} {

   if {$fatal != 0} {
       DLG:error . .error $message "Close window"
       tkWWWNavigateCloseThisWindow
   } {
       DLG:error . .error $message "OK"
   }
}

proc tkWWWOutputMakeText {w} {
    frame $w -relief flat -borderwidth 5
    frame $w.graphics
    pack append $w \
	[ text $w.msg \
	 -yscrollcommand { .vs set } \
	     -width 80 \
	     -height 40 ] { expand fill padx 2 pady 2 }

    return $w
}

proc tkWWWOutputMakeScrollbars {} {
    pack before .f.msg \
	[scrollbar .vs \
	 -command { .f.msg yview } -orient vertical] { right filly }
}

proc tkWWWOutputClearBody {} {
    .f.msg delete 1.0 end
}

proc tkWWWOutputSaveFileAs {file_name} {
    global fsBox

    tkWWWOutputClearBody
    set name [file tail [lindex [tkWWWServerSend HtGetAddressAndTitle] 0]]
    set fsBox(path) ""
    FSBox {Save File} $name \
	"exec mv $file_name \$fsBox(path)/\$fsBox(name)"
    DLG:show . .fsBox
}

