## pan.tcl: Procedures for handling Xmosaic personal annotations
## ===============
## Copyright (C) 1993 Joseph Wang
## See the file COPYRIGHT for conditions
##
## This file is a tcl translation of the procedures
## used in X Mosaic by Marc Anderssen

set tkWWWPanLogFormatCookieOne \
 "ncsa-mosaic-personal-annotation-log-format-1" 
set tkWWWPanAnnotationFormatOne "<ncsa-annotation-format-1>"

set tkWWWPanDefaultDirectory "~/.mosaic-personal-annotations"
set tkWWWPanLogFilename "LOG"
set tkWWWPanAnnotationPrefix "PAN-"


## The format of the list
global tkWWWPanAnnotationList

## Initialize
proc tkWWWPanInitialize {} {
    global tkWWWPanDefaultDirectory 
    global tkWWWPanLogFilename

    tkWWWPanEnsureDirectoryExists $tkWWWPanDefaultDirectory
    tkWWWPanReadLog [glob $tkWWWPanDefaultDirectory/$tkWWWPanLogFilename]
}

## Read a Pan File
proc tkWWWPanReadLog {filename} {
    global tkWWWPanLogFormatCookieOne 
    global tkWWWPanAnnotationList

    if {[catch "open $filename r" file] != 0} {
	return
    }
    
    gets $file line
    if {[string first $tkWWWPanLogFormatCookieOne $line] == -1} {
	return
    }

    gets $file line

    catch {unset tkWWWPanAnnotationList}
    while {[gets $file line] != -1} {
	set tkWWWPanAnnotationList([lindex [split $line] 0]) \
	    [lrange $line 1 end]
    }
    close $file
}

proc tkWWWPanEnsureDirectoryExists {directory} {
    if {![file isdirectory [glob $directory]]} {
	exec mkdir [glob $directory ]
    }
}

proc tkWWWPanIsEditablePan {url} {
    global tkWWWAnnotationFormatOne
    if {[regexp {PAN-[0-9]+\.html} $url] == 1} {
	return 1
    } {
	return 0
    }
}

proc tkWWWPanWriteLog {filename} {
    global tkWWWPanLogFormatCookieOne
    global tkWWWPanAnnotationList
    if [catch {open $filename "w"} file] {
	return
    }

    puts $file $tkWWWPanLogFormatCookieOne
    puts $file "Personal"
    foreach item [array names tkWWWPanAnnotationList] {
	puts $file "$item " nonewline
	foreach piece $tkWWWPanAnnotationList($item) {
	    puts $file $piece nonewline
	    puts $file " " nonewline
	}
	puts $file ""
    }

    close $file
}

proc tkWWWPanWritePan {filename title author text} {
    global tkWWWPanAnnotationFormatOne
    if [catch "open $filename w" file] {
	return
    }
    puts $file "\
$tkWWWPanAnnotationFormatOne\n\
<title>$title</title>\n\
<h1>$title</h1>\n\
<address>$author</address>\n\
<address>[exec date]</address>\n\
_______________________________________\n\
<pre>\n\
$text"
    close $file
}

proc tkWWWPanNewPan {url title author text} {
    global tkWWWPanAnnotationList
    global tkWWWPanDefaultDirectory
    global tkWWWPanLogFilename

    if {[catch {set tkWWWPanAnnotationList($url)} list]} {
	set list ""
    }

# Scan through the files for an available ID
    for {set id 1} {[file exists [tkWWWPanIdToFilename $id]]} {incr id 1} {}
 
    lappend list $id

    set tkWWWPanAnnotationList($url) $list

    tkWWWPanWriteLog  [glob $tkWWWPanDefaultDirectory/$tkWWWPanLogFilename]
    tkWWWPanWritePan [tkWWWPanIdToFilename $id] $title $author $text
}

proc tkWWWPanDeletePan {id} {
    global tkWWWPanAnnotationList
    foreach i [array names tkWWWPanAnnotationList] {
	set j [lsearch $tkWWWPanAnnotationList($i) $id]
	if {j != -1} {
	    set $tkWWWPanAnnotationList($i) \
		[ lreplace $tkWWWPanAnnotationList($i) $j ]
	    if {[llength $tkWWWPanAnnotationList($i)] == 0} {
		unset $tkWWWPanAnnotationList($i)
	    }
	    exec mv [tkWWWPanIdToFilename $id] \
		[format "%s/#%s%d" \
		 $tkWWWPanDefaultDirectory $tkWWWPanLogFilename $id]
	    break
	}
    }
}

proc tkWWWPanModifyPan {id title author text} {
    tkWWWPanWritePan [tkWWWPanIdToFilename $id] $title $author $text
}

proc tkWWWPanIdToFilename {id} {
    global tkWWWPanDefaultDirectory
    global tkWWWPanAnnotationPrefix

    return [glob "$tkWWWPanDefaultDirectory/$tkWWWPanAnnotationPrefix$id.html"]
}

# The next procedure returns a list of hypertext commands which 
# display the annotations
proc tkWWWPanDisplayAnnotations {url} {
    global tkWWWPanAnnotationList
    catch {destroy .menu.annotate.m.annotate}
    menu .menu.annotate.m.annotate
    if {[catch  {set tkWWWPanAnnotationList($url)} list] == 0} {
	foreach item $list {
	    set file [open [tkWWWPanIdToFilename $item] "r"]
	    gets $file line
	    gets $file title
	    gets $file line
	    gets $file line
	    gets $file date

	    regexp {<title>(.*)</title>} $title {} title
	    regexp {<address>(.*)</address>} $date {} date
	    .menu.annotate.m.annotate add command -label "$title $date" \
		-command "tkWWWNavigateRecordAndGoto file:[tkWWWPanIdToFilename $item]"
	}
	return 1
    }
    return 0
}

# return list: {id title author text}

proc tkWWWPanGrokPanPieces {url} {
    global tkWWWPanAnnotationPrefix

    regexp {:(//[^/])*(.*)} $url {} {} filename
    regexp {([0-9]+)\.html} $filename {} id

    if {[catch "open $filename r" file]} {
	return
    }
    gets $file line
    gets $file title
    gets $file line
    gets $file author
    gets $file date
    gets $file line
    gets $file line

    regexp {<address>(.*)</address>} $author {} author
    regexp {<title>(.*)</title>} $title {} title
    regexp {<address>(.*)</address>} $date {} date

    set text [read $file]
    close $file
    return [list $id $title $author $text]
}

