## server.tcl - Communication with the tkWWWServer
## ==============
## Copyright (C) 1992 Joseph Wang
## See the file COPYRIGHT for conditions

## Conventions:
##   all globally visible objects in this file should begin with
##     tkWWWServer
##   all interaction with the tkWWW server should occur through the
##     procedures in this file

## Global Variables

# the file handle of the server
set tkWWWServer {}

# This is the prompt that the server looks for
set tkWWWServerPrompt {}

## *************************
## Procedures for communicating with the TkWWW server
##    tkWWWServerOpen - opens a connection with the tkWWW server and
##       returns the opening message
##    tkWWWServerSend - sends a string to the tkWWW server
##    tkWWWServerReceive - receives a string from the tkWWW server
##    tkWWWServerClose - shuts down the tkWWW server
## *************************

## tkWWWServerOpen: Open connection with tkWWWServer
## ----------
## On entry,
##   exec_string: string to execute
## On exit,
##   the introductory message given by the server
## Side effects,
##   resets the variable tkWWWServer

## Remember that tcl returns value of the the last statement executed in 
## a procedure

proc tkWWWServerOpen {exec_string prompt} {
   global tkWWWServer
   global tkWWWServerPrompt
   # open connection with server
   # There should be some way of figuring out if the server responds
   set tkWWWServer [open "| $exec_string" w+]
   set tkWWWServerPrompt $prompt
   tkWWWServerReceive
}

## tkWWWServerSend: Send message to tkWWW Server
## ----------
## On entry,
##   w: message to server
## On exit,
##   returns response from server
## WARNING:
##   SERVER MUST MAKE SOME SORT OF RESPONSE OR ELSE THERE WILL BE DEADLOCK

proc tkWWWServerSend {w} {
   global tkWWWServer
   puts $tkWWWServer $w
   tkWWWServerReceive
}

## tkWWWServerReceive: Wait for response from server
## ----------
## On exit,
##   returns response from server
## WARNING:
##   SERVER MUST MAKE SOME SORT OF RESPONSE OR ELSE THERE WILL BE DEADLOCK

proc tkWWWServerReceive {} {
  global tkWWWServer
  global tkWWWServerPrompt

  set return_string ""

  # *** THIS FLUSH IS VERY IMPORTANT
  # *** If this isn't put here, you are likely to send the application
  # ***    into deadlock

  flush $tkWWWServer

  # Get stuff from the server until the prompt string is detected
  # Then return everything.  WARNING.  This will freeze if the prompt
  # string is never detected

  while {1} {
      if {[gets $tkWWWServer string] == -1} {
	  tkWWWServerError
      }
      if [regexp $tkWWWServerPrompt $string] {
	  break;
      }
      append return_string $string
      append return_string "\n"
  }

  # Make sure you remove newlines from the end of return string.
  # Otherwise, strange things will happen.
  # 
  # What will happen is this when you try to evaluate
  #   tkWWWGoto [tkWWWParseName $page]
  # tkWWWParseName calls this procedure which returns with 
  #   file:filename.html\n\n
  # When tkWWWGoto then calls this procedure with that as an arg
  # the second newline will cause two "tkWWW:" prompts to appear
  # messing up the sychronoization between the GUI and the backend.

  string trim $return_string
}

## tkWWWServerClose: Close connection with server
## ----------
## Side effects:
##   Closes the pipe to the server, resets the server value

proc tkWWWServerClose {} {
    global tkWWWServer
    if {$tkWWWServer != ""} {
	# Tell the server to exit
	puts $tkWWWServer "exit"
	# Make sure the message gets there
	flush $tkWWWServer
	# Close the pipe
	close $tkWWWServer
	# reset the server value
	set tkWWWServer ""
    }
}

proc tkWWWServerError {} {
    global tkWWWServer
    set tkWWWServer {}
    tkWWWOutputError "Server process no longer exits" 1
}
